from typing import Any, List, Optional, Type, TypeVar
from sqlalchemy.orm import Session

from models import BaseModel, Policy, Coverage, Event, EventHistory

ModelType = TypeVar("ModelType", bound=BaseModel)


class BaseContext:
    """BaseContext object with default methods to Create, Read, Update,
    Delete (CRUD).

    **Parameters**

    * `model`: A SQLAlchemy model class
    * `schema`: A Pydantic model (schema) class
    """

    model: Type[ModelType] = None

    def get_one(self, db: Session, model_id: Any) -> Optional[ModelType]:
        return db.query(self.model).get(model_id)

    def insert(self, db: Session, created_obj: dict) -> Optional[ModelType]:
        created_obj = self.model(**created_obj)
        db.add(created_obj)
        db.commit()
        db.refresh(created_obj)

        return created_obj

    def bulk_insert(self, db: Session, objects):
        db.bulk_save_objects(objects, return_defaults=True)
        db.commit()

        return objects

    def update(self, db: Session, updated_obj: ModelType, data: dict) -> Optional[ModelType]:
        """
        Parameters:
            :param db:
            :param updated_obj: an object need to update.
            :param data: new data
        """
        for field in updated_obj.get_fields():
            if field in data:
                setattr(updated_obj, field, data[field])

        db.add(updated_obj)
        db.commit()
        db.refresh(updated_obj)
        return updated_obj

    def delete(self, db: Session, deleted_obj):
        db.delete(deleted_obj)
        db.commit()
        return deleted_obj


class PolicyContext(BaseContext):
    model = Policy


class EventContext(BaseContext):
    model = Event


class EventHistoryContext(BaseContext):
    model = EventHistory


class CoverageContext(BaseContext):
    model = Coverage

    def get_all(self, db: Session, policy_guid: str) -> List[Coverage]:
        return db.query(self.model).filter(Coverage.policy_guid == policy_guid).all()
