from typing import Any, Dict, List
from sqlalchemy import inspect, Column, String, Date, Integer, DateTime, JSON
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import ForeignKey
from sqlalchemy.ext.declarative import as_declarative, declared_attr
from sqlalchemy.sql import func
import sqlalchemy.dialects.postgresql as postgresql
from helpers import to_snake_case, generate_uuid

policy_schema = {"schema": "policy"}
event_schema = {"schema": "event"}
code_schema = {"schema": "code"}


@as_declarative()
class BaseModel:
    """
    ORM base class
    Following # https://www.sqlstyle.guide/#:~:text=Names%20must%20begin%20with%20a,(first%20name%20becomes%20first_name%20).
    """

    __name__: str

    # Generate __tablename__ automatically
    @declared_attr
    def __tablename__(self) -> str:
        return to_snake_case(self.__name__)

    def get_fields(self) -> List:
        return [c.key for c in inspect(self).mapper.column_attrs]

    def as_dict(self) -> Dict[str, Any]:
        return {field: getattr(self, field) for field in self.get_fields()}


class Policy(BaseModel):
    """
    Put __table_args__ on each model with its schema
    """

    __table_args__ = policy_schema

    policy_guid = Column(String(36), primary_key=True, default=generate_uuid)
    jurisdiction_id = Column(String(50))
    manufacturer_id = Column(String(50))
    plan_guid = Column(String(50))
    policy_number = Column(String(50))
    effective_date = Column(String(50))
    termination_date = Column(Date)
    status_id = Column(String(50))
    deletion_date = Column(Date)


class Event(BaseModel):
    __table_args__ = event_schema

    event_guid = Column(String(36), primary_key=True, default=generate_uuid)
    event_name = Column(String(150))
    event_group_guid = Column(String(36))
    event_order = Column(Integer)
    entity_type = Column(String(10))
    entity_guid = Column(String(36))
    effective_date = Column(Date)
    parameters = Column(JSON)
    rule_name = Column(String)
    created_at = Column(DateTime(timezone=True), default=func.now())
    modified_at = Column(DateTime(timezone=True), default=func.now(), onupdate=func.now())

    histories = relationship("EventHistory", cascade="all,delete", backref="event")

    def __repr__(self):
        return str(self.event_name)


class EventHistory(BaseModel):
    __table_args__ = event_schema

    event_history_guid = Column(String(36), primary_key=True, default=generate_uuid)
    event_guid = Column(String(36), ForeignKey(Event.event_guid, ondelete="CASCADE"))

    parameters = Column(JSON)
    response = Column(JSON)
    status_id = Column(Integer, ForeignKey("event.status_id.status_id"), default=1)
    status = relationship("StatusId", uselist=False)

    created_at = Column(DateTime(timezone=True), default=func.now())

    def __repr__(self):
        return str(self.event_history_guid)


class StatusId(BaseModel):
    __table_args__ = event_schema

    status_id = Column(Integer, primary_key=True, index=True)
    name = Column(String(36))


class Coverage(BaseModel):
    __table_args__ = policy_schema

    coverage_guid = Column(postgresql.UUID, primary_key=True, unique=True)
    application_guid = Column(postgresql.UUID)
    policy_guid = Column(postgresql.UUID)
    coverage_type_id = Column(Integer)
    coverage_level_id = Column(Integer)
    parent_coverage_guid = Column(postgresql.UUID)
    product_guid = Column(postgresql.UUID)
    effective_date = Column(Date)
    termination_date = Column(Date)
    termination_reason_id = Column(String(50))
    status_id = Column(String(50))
    deletion_date = Column(Date)


class CoverageType(BaseModel):
    __table_args__ = {"schema": "code"}
    coverage_type_id = Column(Integer, primary_key=True, index=True)
    Name = Column(String(150))


class CoverageLevel(BaseModel):
    __table_args__ = {"schema": "code"}
    coverage_level_id = Column(Integer, primary_key=True, index=True)
    Name = Column(String(150))
